case class State[S,A](run: S => (A,S)):
  def flatMap[B](f: A => State[S,B]): State[S,B] = State { s =>      
    val (a,s2) = run(s)
    f(a).run(s2)    
  }
  
  def map[B](f: A => B): State[S,B] = 
    flatMap(a => State(s => (f(a),s)))

  def eval(s: S): A = run(s)._1
  def exec(s: S): S = run(s)._2  

object State:
  def get[S]: State[S,S] = new State(s => (s,s))
  def put[S](s_ : S): State[S,Unit] = new State(s => ((),s_))
  def modify[S](f: S => S): State[S,Unit] = get.flatMap(x => put(f(x)))

object Example:
  def push[A](x: A): State[List[A],Unit] = State.modify(x :: _)

  def pop[A]: State[List[A],A] = for
    xs <- State.get
    _ <- State.put(xs.tail)
  yield xs.head
  
  def example: State[List[Int],Int] = for
    _ <- push(10)
    _ <- push(20)
    a <- pop
    b <- pop
    _ <- push (a+b)
    _ <- push(30)
    a <- pop
    b <- pop
    _ <- push (a*b)
    result <- pop
  yield result